#include <stdio.h>
#include <math.h>
#include "clock.h"
#include "sum.h"

#ifndef TSIZE
#define TSIZE 32*1024
#endif
#ifndef NBREPEAT
#define NBREPEAT 100
#endif

inline void dynarray_append (dynarray *a, element *e) {
  if (a->size == a->capacity) {
    if (a->capacity == 0) a->capacity = 1;
    else a->capacity = 2 * a->capacity;
    a->array = realloc(a->array, a->capacity*sizeof(element));
  }      
  a->array[a->size] = *e;
  (a->size)++;
}

void alloc_fill_array(element **parray, int size) {
  int     i;
  element fval;
  element *array;
  array = (element*) valloc(size*sizeof(element));
  for (i=0; i<size; i++) {
    fval = .1;
    array[i] = fval;
  }
  *parray = array;
}

void alloc_fill_dynarray(dynarray *darray, int size) {
  int     i;
  element fval;
  for (i=0; i<size; i++) {
    fval = .1;
    dynarray_append(darray, &fval); 
  }
}

element repeat_sum_dynarray(dynarray *darray, int size, int repeat) {
  int k;
  element sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += sum_dynarray(darray, size);
  }
#pragma omp barrier
  return sumval;
}

element repeat_sum_array(element *array, int size, int repeat) {
  int k;
  element sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += sum_array(array, size);
  }
#pragma omp barrier
  return sumval;
}

element repeat_sum_rand_array(element *array, int size, int repeat) {
  int k;
  element sumval = 0;
#pragma omp barrier
  for (k=0; k<repeat; k++) {
    sumval += sum_rand_array(array, size);
  }
#pragma omp barrier
  return sumval;
}


/* Program that tests dynamic array */
int main(int argc, char **argv) {
  const long long int Nmax    = TSIZE;
  element sumval_dyn = 0, sumval_stat = 0, sumval_rand = 0;
#pragma omp parallel reduction(+:sumval_stat,sumval_dyn,sumval_rand) 
  {
    dynarray  darray = {0};
    long long int NbBytes;
    long long int timesNb;
    int size;
    element *array;
    char message[50];
    GETTIME(alloc_fill_dynarray(&darray, Nmax),"fill_dynarray");
    GETTIME(alloc_fill_array(&array, Nmax),"fill_array");
    timesNb = NBREPEAT;
    size    = Nmax;
    for (; size >= 128; size/=4, timesNb*=4) {  
      NbBytes = size*sizeof(element)*timesNb;
      sprintf(message,"sum_array, size=%d",size);
      GETBWIDTH(sumval_stat=repeat_sum_array(array, size, timesNb),
		NbBytes,message);
      sprintf(message,"sum_dynarray, size=%d",size);
      GETBWIDTH(sumval_dyn=repeat_sum_dynarray(&darray, size, timesNb),
		NbBytes,message);
      sprintf(message,"sum_randarray, size=%d",size);
      GETBWIDTH(sumval_rand=repeat_sum_rand_array(array, size, timesNb),
		NbBytes,message);
    }
  }
  printf("sums = %f %f %f\n",sumval_dyn, sumval_stat,sumval_rand);
  return (0);
}
