#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "clock.h"
#include "haar.h"
#ifndef DRATE
#define DRATE 1
#endif
#ifndef TSIZE
#define TSIZE 1024*1024
#endif

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void print1d(double *data, int size, double dx, char *fname) {
  FILE*  fd;
  register int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    val = data[i];
    fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "data" except NOTAVUE values. */
/* The name of the file is "fname", the "size" specifies the array size */
void wprint1d(double *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    val = data[i];
    if (val != NOTAVALUE)
      fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "FTYPE *data" except NOTAVUE values. */
/* The name of the file is "fname", the "size" specifies the array size */
void wbprint1d(FTYPE *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    _GETF(data,i,val);
    if (val != NOTAVALUE)
      fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "FTYPE* data". */
/* The name of the file is "fname", the "size" specifies the array size */
void waprint1d(FTYPE *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    _GETF(data,i,val);
      fprintf(fd,"%f %.3e\n",dx*i,val);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void average1d(double *data, int size, char *mess) {
  register int i;
  register double sum;
  sum = 0;
  for (i = 0; i < size; i++) {
    sum += data[i];
  }
  printf("%s %e\n",mess,sum/(float)size);
}

/* Initialize the initial array with a sine function 
   multiplied by a decreasing exp(-x) coefficient */
  /* Retrieve the value of 2*PI */
void init_func1(double *data, int size) {
  double  dpi, coeff, dx, val, maxval, inv_maxval;
  int     modex, i;
  /* Two times PI */
  dpi   = 8*atan(1.);
  /* Excited mode */
  modex = 6;
  /* Mesh size */
  dx    = dpi/size;
  coeff = dx*modex;
  /* Initialize random generator */
  srand48(4);
  /* Fill "data" array with a signal + noise */
  /* The signal is a sine times an exponentional decrease */
  maxval = 0;
  for (i = 0; i < size; i++) {
    val = exp(-dx*2*abs(i-size*.5))*sin(coeff*i) + .005*drand48();
    data[i] = val;
    val = fabs(val);
    maxval = (val > maxval?val:maxval);
  }
  inv_maxval = 1./maxval;
  for (i = 0; i < size; i++) data[i] = data[i] * inv_maxval;
}

/* Program that runs a haar wavelet scheme
   on a 1D data set : a noisy sine signal */
int main(int argc, char **argv) {
  double  *f_old, *f_wav, *f_tmp;
  FTYPE   *s_old, *s_new, *s_wav, *swap;
  int     Nmax = TSIZE;
  int     nnztot;
  double  Lp = 1000;
  double  normcoef = pow(2,-1/Lp);
  double  thresholdMax = .009;
  int     i;
  double  displacement;
  double  dx;
  int     nbdisplacements;
  char    message[50];

  f_old = (double*)malloc(Nmax*sizeof(double));
  f_wav = (double*)malloc(Nmax*sizeof(double));
  f_tmp = (double*)malloc(Nmax*sizeof(double));
  _INITF (&s_new,Nmax);
  _INITF (&s_wav,Nmax);
  _INITF (&s_old,Nmax);
  printf("Nmax = %d\n",Nmax);

  dx    = atan(8.)/Nmax;
  /* Initialize initial array with a modified sine signal */
  GETTIME(init_func1(f_old,Nmax),"init_func1:");
  nbdisplacements = 32;
  displacement = Nmax/nbdisplacements;
  /* print input signal into a file 'signal' */
  WAVDEBUG(print1d(f_old,Nmax,dx,"input.dat"));
  average1d(f_old,Nmax,"average of input:");
  /* Copy signal from f_old to f_wav */
  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
  /* Wavelet forward transform */
  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
  /* Thresholding of details */
  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,f_old),"threshold:");
  wprint1d(f_old,Nmax,dx,"ref_wdeb.dat");

  /* Copy signal from f_wav to f_new */
  GETTIME(wav_copyfunc_as(f_wav,s_new,s_wav,Nmax),"fcopy_B:");
  /* Wavelet inverse transform */
  GETTIME(wav_adapt_itransform(s_new,Nmax),"itransf:");

  WAVDEBUG(waprint1d(s_new,Nmax,dx,"ref_out1.dat"));

  /* Copy signal from f_old to s_old */
  GETTIME(wav_copyfunc_as(f_old,s_old,s_wav,Nmax),"fcopy_C:");
  /* Haar special inverse transform */
  GETTIME(wav_specf_itransform(s_old,Nmax,f_tmp),"specf_itransf:");
  average1d(f_tmp,Nmax,"average of input:");
  WAVDEBUG(print1d(f_tmp,Nmax,dx,"ref_out2.dat"));
  sprintf(message,"out_%d.dat",0);
  printf("out %s\n",message);
  WAVDEBUG(print1d(f_tmp,Nmax,dx,message));

  for (i = 1; i <= nbdisplacements; i++) {
    /* Advection */
    GETTIME(wav_advect_grid(s_old,Nmax,-displacement,s_new, s_wav),"advect_grid:");
    GETTIME(wav_advect(s_old,Nmax,-displacement,s_new),"advect");
    GETTIME(nnztot=wav_getnnz(s_new,Nmax),"count nnz:");
    printf("NNZ %d\n",nnztot);
    GETTIME(wav_copyfunc_ss(s_new,s_wav,Nmax),"fcopy_loop:");
    /* Wavelet forward transform */
    GETTIME(wav_adapt_ftransform(s_wav,Nmax),"adapt_ftransf_B:");
    /* Thresholding of details */
    GETTIME(wav_adapt_thresholding(s_wav,Nmax,normcoef,thresholdMax,s_new),"threshold:");
    GETTIME(nnztot=wav_getnnz(s_wav,Nmax),"count nnz:");
    printf("NNZ AT THE END OF ITER %d: %d\n",i,nnztot);
    if (i%DRATE == 0) {
      /* Copy signal from s_new to s_wav */
      WAVDEBUG(GETTIME(wav_copyfunc_ss(s_new,s_wav,Nmax),"fcopy_diag:"));
      /* Haar special inverse transform */
      WAVDEBUG(GETTIME(wav_specf_itransform(s_wav,Nmax,f_tmp),"specf_itransf:"));
      /* print restored signal into a file 'output' */
      sprintf(message,"out_%d.dat",i/DRATE);
      printf("out %s\n",message);
      WAVDEBUG(print1d(f_tmp,Nmax,dx,message));
      WAVDEBUG(average1d(f_tmp,Nmax,"average of output:"));
    }
    swap = s_new;
    s_new = s_old;
    s_old = swap;
  }
  wbprint1d(s_old,Nmax,dx,"ref_wend.dat");
  /* Haar special inverse transform */
  GETTIME(wav_specf_itransform(s_old,Nmax,f_tmp),"specf_itransf:");
  sprintf(message,"out_%d.dat",i/DRATE);
  WAVDEBUG(print1d(f_tmp,Nmax,dx,message));
  
  free(f_old);
  free(f_wav);
  free(f_tmp);
  _DESTROY(&s_old);
  _DESTROY(&s_wav);
  _DESTROY(&s_new);
  return 0;
}
