#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#ifndef HAAR_H
#define HAAR_H

#ifdef DEBUG
#define WAVDEBUG(expr) expr
#else
#define WAVDEBUG(expr) 
#endif

#define NOTAVALUE ((double)1e20)
#define _KEYTYPE uint64_t
#define _VALTYPE double  

#ifdef _HASH

#include "hash.h"
#define FTYPE hashtable

#else /* HASH*/
/* Instead of hash tables use simply dense arrays */
#define FTYPE double
#define    _DESTROY(_phatab) {free(*_phatab); (*_phatab)=NULL; }
#define      _INITF(_phatab,_size)			\
  ((*_phatab)= (FTYPE*) malloc(_size*sizeof(FTYPE)))
#define     _RESETF(_hatab,_size)				\
  {int _i; for (_i=0; _i<_size; _i++) _hatab[_i] = NOTAVALUE;}
#define    _SETF_NEW(_hatab,_key,_val)  _hatab[_key] = _val;
#define _GETPOINTER(_hatab,_key,_pval) _pval=&(_hatab[_key]);
#define       _GETF(_hatab,_key,_val)  _val = _hatab[_key];
#endif /* HASH*/

void wav_zeroing( double *fold, int N );
int wav_log2(int input);
int wav_ilevel(int input, int nblevel);
int wav_level(int input, int nblevel);
int wav_itf(int input, int nblevel);
void wav_getmasks(int nblevel, int *masks);
void wav_getmask(int itc, int *itcmask);
void wav_predict(double* fwav, int itc, int N, int dir );
void wav_ftransform( double *fwav, int N );
void wav_adapt_predict( FTYPE *swav, int itc, int N, int dir );
void wav_adapt_ftransform( FTYPE *swav, int N );
void wav_itransform( double *fwav, int N );
void wav_adapt_itransform( FTYPE *swav, int N );
void wav_reorder_grid( FTYPE *sold, FTYPE *snew, int N );
void wav_thresholding( double *fwav, int N, double normcoef , double threshold, double *fold );
void wav_adapt_thresholding( FTYPE *swav, int N, double normcoef , double threshold);
int wav_getnnz( FTYPE *swav, int N );
void wav_copyfunc_aa( double *fold, double *fnew, int N);
void wav_copyfunc_as( double *fold, FTYPE *snew, FTYPE *stmp, int N);
void wav_copyfunc_ss( FTYPE *sold, FTYPE *snew, int N);
void wav_copyfunc_sa( FTYPE *sold, double *fnew, int N);
void wav_review( FTYPE *sold, int N );
void wav_advect_grid( FTYPE *swav, int N, double displacement, FTYPE *snew, FTYPE *stmp );
void wav_advect( FTYPE *sold, int N, double displacement, FTYPE *snew);

#endif
