#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "haar.h"
#include "clock.h"

#ifndef TSIZE
#define TSIZE 1024
#endif

void print1d(double *data, int size, double dx, char *fname);
void average1d(double *data, int size, char *mess);

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void print1d(double *data, int size, double dx, char *fname) {
  FILE*  fd;
  int i;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (i = 0; i < size; i++) {
    fprintf(fd,"%f %.3e\n",dx*i,data[i]);
  }
  fclose(fd);
}

/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void average1d(double *data, int size, char *mess) {
  int i;
  double sum;
  sum = 0;
  for (i = 0; i < size; i++) sum += data[i];
  printf("%s %e\n",mess,sum/(float)size);
}

/* Program that runs a haar scheme
   on a 1D data set */
int main(int argc, char **argv) {
  double *data;
  double dpi, coeff, dx, ftmp;
  int    Nmax = TSIZE;
  double Lp = 100;
  double normcoef = pow(2,-1/Lp);
  double thresholdMax = .02;
  int    i, modex;
  data = (double*)malloc(Nmax*sizeof(double));
  /* Retrieve the value of 2*PI */
  dpi   = 8*atan(1.);
  modex = 6;
  /* Mesh size */
  dx    = dpi/Nmax;
  coeff = dx*modex;
  /* Initialize the random generator */
  srand48(4);
  /* Fill "data" array with a signal + noise */
  /* The signal is a sine times an exponentional decrease */
  for (i = 0; i < Nmax; i++) {
    ftmp    = exp(-dx*2.*fabs((double)i-(double)Nmax*.5));
    ftmp   *= sin(coeff*i);
    data[i] = ftmp + .005*drand48();
  }
  /* print input signal into a file 'signal' */
  WAVDEBUG(print1d(data,Nmax,dx,"input"));
  average1d(data,Nmax,"average of input:");
  /* Wavelet forward transform */
  //  GETTIME(haar_ftransform(data,Nmax),"ftransfv2:");
  GETTIME(haar_fdwt(data,Nmax),"ftransfv1:");
  /* Thresholding of details */
  GETTIME(haar_thresholding(data,Nmax,normcoef,thresholdMax),"threshold:");
  /* Wavelet inverse transform */
  //  GETTIME(haar_itransform(data,Nmax),"itransfv2:");
  GETTIME(haar_idwt(data,Nmax),"itransfv1:");
  /* print restored signal into a file 'output' */
  WAVDEBUG(print1d(data,Nmax,dx,"output"));
  average1d(data,Nmax,"average of output:");
  free(data);
  return 0;
}
