#include <glib.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include "hash.h"
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <stdlib.h>

#ifndef TSIZE
#define TSIZE 10000000
#endif
#ifndef STRIDE
#define STRIDE 15
#endif

inline void clockget_ (int64_t *timepoint) {
  struct timeval tv2;
  gettimeofday(&tv2, (struct timezone*)0);
  timepoint[0] = (tv2).tv_sec;
  timepoint[1] = (tv2).tv_usec;
}

inline double clockdiff_ (int64_t *tp1, int64_t *tp2) {
  return((tp2[0]-tp1[0])+0.000001*(tp2[1]-tp1[1]));
}


inline void hash_init_table(hashtable **phatab, int nbelts) {
  int i, nblevel;
  hashtable *hatab;
  double fnblevel;
  fnblevel    = log2(((double)nbelts));
  nblevel     = (int)fnblevel;
  hatab       = (hashtable *)   malloc(sizeof(struct shashtable));
  *phatab     = hatab;
  hatab->keys = (HASH_KEYTYPE*) malloc (nbelts*sizeof(HASH_KEYTYPE));
  hatab->vals = (HASH_VALTYPE*) malloc (nbelts*sizeof(HASH_VALTYPE));
  hatab->ends = (int*) malloc((nblevel+2)*sizeof(int));
  hatab->hash = g_hash_table_new(g_int_hash, g_int_equal);
  hatab->maxsize = nbelts;
  hatab->cursize = 0;
  for(i = 0; i < nbelts; i++) {
    //    if (i%(nbelts/100)==0) printf("elt %d size %d \n",i,nbelts);
    hatab->keys[i] = 0;
    hatab->vals[i] = 0;
  }
}

inline void hash_destroy_table(hashtable **phatab) {
  hashtable *hatab;
  hatab = *phatab;
  free(hatab->keys);
  free(hatab->vals);
  g_hash_table_destroy(hatab->hash);
  free(hatab);
  *phatab = NULL;
}


int main(){
  g_thread_init (NULL);
#pragma omp parallel 
  {
  GHashTableIter  iter;
  int64_t         i;
  int64_t         size;
  HASH_KEYTYPE    key;
  HASH_VALTYPE    val;
  HASH_KEYTYPE    *pkey;
  HASH_VALTYPE    *pval;
  double          sum;
  int64_t         t1[2], t2[2];
  hashtable       *htab;

  hash_init_table(&htab,(int)TSIZE);
#pragma omp barrier
  clockget_(t1);
  for(i=0;i<TSIZE;i+=STRIDE) {
    key = i;
    val = i*.1;
    HASH_SETF_NEW(htab,key,val)
  }
  clockget_(t2);
  printf ("hash inserts time    %f\n",clockdiff_(t1,t2)); fflush(stdout);
  HASH_GET_SIZE(htab,size);
  printf ("hash nb elts         %ld\n",(long int)size);

#pragma omp barrier
  clockget_(t1);
  sum = 0.;
  g_hash_table_iter_init (&iter, htab->hash);
  while (g_hash_table_iter_next (&iter, (gpointer) &pkey, (gpointer) &pval)) {
    sum += *pval;
  }
  clockget_(t2);
  printf ("traversal 1 [hash table iterator O(S)] sum result: %.1f , execution time      %f \n",
	  sum,clockdiff_(t1,t2));


#pragma omp barrier
  clockget_(t1);
  sum = 0.;
  for(i=0;i<TSIZE;i++) {
    val = htab->vals[i];
    if (val != NOTAVALUE) {
      sum += val;
    }
  }
  clockget_(t2);
  printf ("traversal 2 [dense array         O(N)] sum result: %.1f , execution time      %f \n",
	  sum,clockdiff_(t1,t2));

#pragma omp barrier
  clockget_(t1);
  sum = 0.;
  for(i=0;i<TSIZE;i++) {
    key = i;
    HASH_GETF(htab,key,val);
    if (val != NOTAVALUE) {
      sum += val;
    }
  }
  clockget_(t2);
  printf ("traversal 3 [hash lookups        O(N)] sum result: %.1f , execution time      %f \n",
	  sum,clockdiff_(t1,t2));

#pragma omp barrier
  clockget_(t1);
  sum = 0.;
  for(i=0;i<TSIZE;i++) {
    key = (3413*i+1607)%TSIZE;
    val = htab->vals[key];
    if (val != NOTAVALUE) {
      sum += val;
    }
  }
  clockget_(t2);
  printf ("traversal 4 [non-seq access      O(N)] sum result: %.1f , execution time      %f \n",
	  sum,clockdiff_(t1,t2));

  }
  return 0;
}

//#pragma omp barrier
//  clockget_(t1);
//  sum = 0.;
//  for (i=0; i<htab->cursize; i++) {
//    sum += htab->vals[i];
//  }
//  clockget_(t2);
//  printf ("traversal 3 [hash table entries]  sum result: %.1f , execution time      %f \n",
//	  sum,clockdiff_(t1,t2));
//  
//
//  

//  clockget_(t1);
//  max = 0.;
//  for(i=0;i<TSIZE;i+=STRIDE) {
//    result = (double*)g_hash_table_lookup(g_hash_table,(void*)&i);
//    max = (fabs(*result)>max?fabs(*result):max);
//  }
//  clockget_(t2);
//  printf ("lookup time          %f\n",clockdiff_(t1,t2));
