#include <glib.h>
#include <math.h>

#define NOTAVALUE (double)1e30

#define HASH_KEYTYPE uint32_t
#define HASH_VALTYPE float
#define HASH_TABTYPE GHashTable

struct shashtable {
  HASH_KEYTYPE *keys;
  HASH_VALTYPE *vals;
  HASH_TABTYPE *hash;
  int   *ends;
  int maxsize;
  int cursize;
};

typedef struct shashtable hashtable;

void hash_init_table(hashtable **phatab, int nbelts);
void hash_destroy_table(hashtable **phatab);

#define _INITF(_phatab,_size) hash_init_table(_phatab,_size)

#define _DESTROY(_phatab) hash_destroy_table(_phatab)

#define _RESETF(_hatab,_size) {\
  g_hash_table_steal_all(_hatab->hash);		\
  _hatab->cursize = 0;				\
  }

#define HASH_GET_SIZE(_hatab,_size) {		\
  _size=g_hash_table_size(_hatab->hash);		\
  }

#define HASH_SETF_NEW(_hatab,_key,_val) {					\
    if (_hatab->cursize-1 >= _hatab->maxsize) {		\
      printf("error out of bound in _SETADDF\n");fflush(stdout);exit(2); \
    }									\
    _hatab->keys[_hatab->cursize] = _key;			\
    _hatab->vals[_hatab->cursize] = _val;			\
    g_hash_table_insert(_hatab->hash,			\
			(gpointer)&(_hatab->keys[_hatab->cursize]),	\
			(gpointer)&(_hatab->vals[_hatab->cursize]));	\
    (_hatab->cursize)++;						\
  }

#define HASH_SETF_MODIFY(_hatab,_key,_val,_ind) {			\
    if (_ind >= _hatab->maxsize) {					\
      printf("error out of bound in _SETINDEXF\n");fflush(stdout);exit(2); \
    }			\
    _hatab->keys[_ind] = _key;						\
    _hatab->vals[_ind] = _val;						\
    g_hash_table_insert(_hatab->hash,					\
			(gpointer)&(_hatab->keys[_ind]),		\
			(gpointer)&(_hatab->vals[_ind]));		\
    _hatab->cursize=(_ind+1>_hatab->cursize?_ind+1:_hatab->cursize);	\
  }

#define HASH_GETPOINTER(_hatab,_key,_pval) {			     \
    _pval=((double*)g_hash_table_lookup(_hatab->hash,(void*)&_key)); \
  }

#define HASH_GETF(_hatab,_key,_val)	{					\
    HASH_VALTYPE* _pval=((double*)g_hash_table_lookup(_hatab->hash,(void*)&_key)); \
    _val=((_pval == NULL)?NOTAVALUE:*(_pval));				\
  }

//    printf("getf %d %x\n",_key, (uint)_pval);fflush(stdout);	

//  g_hash_table_iter_init (&iter, g_hash_table);
//
//  clockget_(t1);
//  max = 0.;
//  while (g_hash_table_iter_next (&iter, (gpointer) &key, (gpointer) &val)) {
//    max = (fabs(*val)>max?fabs(*val):max);
//  }
//  clockget_(t2);
//  printf ("traversal time       %f\n",clockdiff_(t1,t2));
//  
//  clockget_(t1);
//  max = 0.;
//  for(i=0;i<TABLE_SIZE;i++) {
//    result = (double*)g_hash_table_lookup(g_hash_table,(void*)&i);
//    max = (fabs(*result)>max?fabs(*result):max);
//  }
//  clockget_(t2);
//  printf ("lookup time          %f\n",clockdiff_(t1,t2));
