#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "clock.h"


#ifndef TSIZE
#define TSIZE 1024*1024
#endif

#ifdef DEBUG
#define WAVDEBUG(expr) expr
#else
#define WAVDEBUG(expr) 
#endif

struct sinfo {
  int    i;
  double d;
};

typedef struct sinfo info;

struct sllist {
  struct sllist *next;
  struct sllist *prev;
  struct sinfo  data;
};

typedef struct sllist llist;

void sum_list(llist *list) {
  register llist *iterator = NULL;
  register double sum;
  for (sum = 0, iterator = list; iterator; iterator = iterator->next) {
    sum += (iterator->data).d;
  }
  printf("sum_list %f\n",sum);
}

void sum_list_prefetch(llist *list, int prefetch_advance) {
  register llist *iterator = NULL;
  register double sum;
#ifdef PREFETCH
  register llist *ahead;
  int c;
  for (c = 0, ahead = list; c < prefetch_advance; c++) 
    if (ahead != NULL) {
      __builtin_prefetch(ahead->next,0,3);
      ahead = ahead->next;
    }
#endif
  for (sum = 0, iterator = list; iterator; iterator = iterator->next) {
#ifdef PREFETCH
    if (ahead != NULL) {
      __builtin_prefetch(ahead->next,0,3);
      ahead = ahead->next;
    }
#endif
    sum += (iterator->data).d;
  }
  printf("sum_list %f\n",sum);
}

void display_list(llist *list) {
  register llist *iterator = NULL;
  register int i;
  register info *pdata;
  register int div;
  div = TSIZE/8;
  for (i = 0, iterator = list; iterator; iterator = iterator->next, i++) {
    pdata = &(iterator->data);
    if (i % div == 0) 
      printf("%d item %d : %f\n", i, pdata->i, pdata->d);
  }
}

void display_tab(llist *ltab) {
  register info *pdata;
  register int i;
  for (i=0; i<TSIZE; i++) {
    pdata = &(ltab[i].data);
    printf("%d item %d : %f\n", i, pdata->i, pdata->d);
  }
}

void sum_tab(llist *ltab) {
  register double sum;
  register info *pdata;
  register int i;
  sum = 0;
  for (i=0; i<TSIZE; i++) {
    pdata = &(ltab[i].data);
    sum += pdata->d;
  }
  printf("sum_tab %f\n",sum);
}

struct keyindex {
  int   key;
  int   ind;
};

static int int_compare(const void* a, const void* b) {
  struct keyindex *gda = (struct keyindex *) a;
  struct keyindex *gdb = (struct keyindex *) b;
  return (gda->key > gdb->key);
}
/* store the array in a list-like
   data structure, to get the same
   overall data size */
void build_tab(llist **ptab) {
  llist *ltab = NULL;
  int i;
  info *pdata;
  srand48(2);
  ltab = (llist *) malloc(TSIZE * sizeof(struct sllist));
  for (i = 0; i < TSIZE; i++) {
    //add data to list
    pdata = &(ltab[i].data);
    pdata->i = lrand48();
    pdata->d = i*.1;
    ltab[i].prev = NULL;
    ltab[i].next = NULL;
  }
  *ptab  = ltab;
}

void build_links(llist **plist, llist **ptab) {
  struct keyindex *ki = NULL;
  llist *ltab = NULL;
  llist *list, *next;
  int i;
  ltab = *ptab;
  ki = (struct keyindex *) malloc(TSIZE * sizeof(struct keyindex));
  for (i = 0; i < TSIZE; i++) {
    ki[i].ind = i;
    ki[i].key = ltab[i].data.i;
  }
  qsort (ki,TSIZE,sizeof(struct keyindex),int_compare);
  list = &(ltab[ki[0].ind]);
  list->prev = NULL;
  *plist = list;
  for (i = 1; i < TSIZE; i++) {
    next = &(ltab[ki[i].ind]);
    list->next = next;
    next->prev = list;
    list = next;
  }
  list->next = NULL;
  free(ki);
}

void list_copy_sort(llist **plist, llist **ptab) {
  llist *newtab = NULL;
  llist *list;
  int i;
  newtab = (llist *) malloc(TSIZE * sizeof(struct sllist));
  list = *plist;
  newtab[0].data = list->data;
  newtab[0].prev = NULL;
  for (i = 1; i < TSIZE; i++) {
    list = list->next;
    newtab[i-1].next = &(newtab[i]);
    newtab[i].data = list->data;
    newtab[i].prev = &(newtab[i-1]);
  }
  newtab[TSIZE-1].next = NULL;
  *plist = newtab;
  free(*ptab);
  *ptab = newtab;
}




int main(int argc, char *argv[]) {
#pragma omp parallel
  {
    llist *list = NULL;
    llist *dtab = NULL;
    char mess[100];
#ifdef PREFETCH
    int i;
#endif

#pragma omp barrier
    printf ("SIZE (linked-list or array) %.6f MB\n",
	    1.* (TSIZE * sizeof(llist))/(1024*1024));
    build_tab (&dtab);
    //  GETTIME(display_tab(dtab),"TIME display_list");
    build_links (&list,&dtab);
#pragma omp barrier
    WAVDEBUG(GETTIME(display_list(list),"TIME display_list"));
#pragma omp barrier
    GETTIME(sum_list(list),
	    "TIME Initial traversal (linked list)      ");
#ifdef PREFETCH
    for (i = 0; i<=300; i+=100) {
      sprintf(mess,
	      "TIME prefetch traversal (parameter %6d)",i);
#pragma omp barrier
      GETTIME(sum_list_prefetch(list,i),mess);
    }
#endif
    //print the list data
#pragma omp barrier
    GETTIME(sum_tab(dtab),
	    "TIME dynamic array traversal              ");
#pragma omp barrier
    GETTIME(list_copy_sort(&list, &dtab),
	    "TIME sort list in memory                  ");
#pragma omp barrier
    GETTIME(sum_list(list),
	    "TIME copy/reorder traversal               ");
#pragma omp barrier
    WAVDEBUG(GETTIME(display_list(list),
		     "TIME display_list                         "));
    //free the list and its records
    free(dtab);
  }
  return 0;
} 

