#include <glib.h>
#include <math.h>
#include "sstruct.h"

gboolean return_true(gpointer k, gpointer v, gpointer p) {
  return TRUE;
}

#ifdef GUINT_LONG
/* (sqrt(5) - 1) * 2^63 */
#define HASH_CONST 11400714819323198485
#else
/* (sqrt(5) - 1) * 2^31 */
#define HASH_CONST 2654435769
#endif


guint
hashfunc_knuth (gconstpointer v) {
  // Multiply by the golden ratio 
  // (Knuth, Art of Computer Programming, Vol 3, Section 6.4
  return HASH_CONST* (*(const guint*) v);
}

guint
hashfunc_knuth_division (gconstpointer v) {
  guint a = *(const guint*) v;
  a = a*(a+3);
  return a;
}

guint 
hashfunc_jenkins(gconstpointer v) {
  guint register a = *(const guint*) v;
  a = (a+0x7ed55d16) + (a<<12);
  a = (a^0xc761c23c) ^ (a>>19);
  a = (a+0x165667b1) + (a<<5);
  a = (a+0xd3a2646c) ^ (a<<9);
  a = (a+0xfd7046c5) + (a<<3);
  a = (a^0xb55a4f09) ^ (a>>16);
  return a;
}


void hash_init_table(hashtable **phatab, int N) {
  int i, nblevel;
  hashtable *hatab;
  double fnblevel;
  int maxi = 1;
  fnblevel    = log2(((double)N));
  nblevel     = (int)fnblevel;
  hatab       = (hashtable *) malloc(sizeof(struct shashtable));
  *phatab     = hatab;
  hatab->keys = (_KEYTYPE*) malloc (maxi*sizeof(_KEYTYPE));
  hatab->vals = (_VALTYPE*) malloc (maxi*sizeof(_VALTYPE));
  hatab->ends = (int*) malloc((nblevel+2)*sizeof(int));
  hatab->hash = g_hash_table_new(hashfunc_knuth, g_int_equal);
  hatab->maxkeys = maxi;
  hatab->curkeys = 0;
  for(i = 0; i < maxi; i++) {
    //    if (i%(nbelts/100)==0) printf("elt %d size %d \n",i,nbelts);
    hatab->keys[i] = 0;
    hatab->vals[i] = FORBIDDEN;
  }
}

void hash_destroy_table(hashtable **phatab) {
  hashtable *hatab;
  hatab = *phatab;
  free(hatab->keys);
  free(hatab->vals);
  g_hash_table_destroy(hatab->hash);
  free(hatab);
  *phatab = NULL;
}

