#ifndef HASH_H
#define HASH_H

#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>


#include <glib.h>

#define _TABTYPE GHashTable

struct shashtable {
  _KEYTYPE *keys;
  _VALTYPE *vals;
  _TABTYPE *hash;
  int   *ends;
  int maxkeys;
  int curkeys;
};

typedef struct shashtable hashtable;

gboolean return_true(gpointer k, gpointer v, gpointer p);
void hash_init_table(hashtable **phatab, int nbelts);
void hash_destroy_table(hashtable **phatab);

#define _INITF(_phatab,_noparam,_size) hash_init_table(_phatab,_size)

#define _DESTROY(_phatab) hash_destroy_table(_phatab)

//NEW feature: g_hash_table_steal_all(_hatab->hash);		

#define _GET_SIZE(_hatab,_size) {			\
    _size=g_hash_table_size(_hatab->hash);		\
  }

#define _RESETF(_hatab,_size) {  \
  int _mysize; \
  _GET_SIZE(_hatab,_mysize); \
  if (_mysize > 0) g_hash_table_foreach_steal(_hatab->hash, (GHRFunc)return_true, (gpointer)NULL); \
  _hatab->curkeys = 0;				\
  }

#define _SETF_NEW(_hatab,_keyx,_keyy,_N,_val) {				\
    if (_hatab->curkeys == _hatab->maxkeys) {				\
      int _i, _oldcurkeys, _maxi = _hatab->maxkeys*2;			\
      _oldcurkeys = _hatab->curkeys;                                    \
      _RESETF(_hatab,0);                                                \
      _hatab->curkeys = _oldcurkeys;                                    \
      _hatab->keys = (_KEYTYPE*) realloc (_hatab->keys,_maxi*sizeof(_KEYTYPE)); \
      _hatab->vals = (_VALTYPE*) realloc (_hatab->vals,_maxi*sizeof(_VALTYPE)); \
      _hatab->maxkeys = _maxi;						\
      for(_i=_hatab->curkeys; _i< _hatab->maxkeys;_i++) {               \
         _hatab->keys[_i] = 0;                                          \
         _hatab->vals[_i] = FORBIDDEN;                                  \
      }                                                                 \
      for(_i=0; _i<_hatab->curkeys; _i++)                               \
          g_hash_table_insert(_hatab->hash,			        \
		(gpointer)&(_hatab->keys[_i]),                    	\
		(gpointer)&(_hatab->vals[_i]));           	        \
    }              							\
    _hatab->keys[_hatab->curkeys] = _KEYXY(_keyx,_keyy);		\
    _hatab->vals[_hatab->curkeys] = _val;			        \
    g_hash_table_insert(_hatab->hash,			                \
			(gpointer)&(_hatab->keys[_hatab->curkeys]),	\
			(gpointer)&(_hatab->vals[_hatab->curkeys]));	\
    (_hatab->curkeys)++;						\
  }

#define _SETF_MODIFY(_hatab,_keyx,_keyy,_val,_ind) {				\
    if (_ind >= _hatab->maxkeys) {					\
      int _i,_oldcurkeys,_maxi = _ind*2;						\
      _hatab->keys = (_KEYTYPE*) realloc (_hatab->keys,_maxi*sizeof(_KEYTYPE)); \
      _hatab->vals = (_VALTYPE*) realloc (_hatab->vals,_maxi*sizeof(_VALTYPE)); \
      _hatab->maxkeys = _maxi;						\
      _oldcurkeys = _hatab->curkeys;                                    \
      _RESETF(_hatab,0);                                                \
      _hatab->curkeys = _oldcurkeys;                                    \
      for(_i=_hatab->curkeys; _i< _hatab->maxkeys;_i++) {               \
         _hatab->keys[_i] = 0;                                          \
         _hatab->vals[_i] = FORBIDDEN;                                  \
      }                                                                 \
      for(_i=0; _i<_hatab->curkeys;_i++)                                \
         if (_hatab->vals[_i] != FORBIDDEN)                             \
             g_hash_table_insert(_hatab->hash,			        \
		(gpointer)&(_hatab->keys[_i]),                    	\
		(gpointer)&(_hatab->vals[_i]));           	        \
    }									\
    _hatab->keys[_ind] = _KEYXY(_keyx,_keyy);				\
    _hatab->vals[_ind] = _val;						\
    g_hash_table_insert(_hatab->hash,					\
			(gpointer)&(_hatab->keys[_ind]),		\
			(gpointer)&(_hatab->vals[_ind]));		\
    _hatab->curkeys=(_ind+1>_hatab->curkeys?_ind+1:_hatab->curkeys);	\
  }

#define _SETF_MODIFY_K(_hatab,_key,_val,_ind) {				\
    if (_ind >= _hatab->maxkeys) {					\
      int _i,_oldcurkeys,_maxi = _ind*2;						\
      _hatab->keys = (_KEYTYPE*) realloc (_hatab->keys,_maxi*sizeof(_KEYTYPE)); \
      _hatab->vals = (_VALTYPE*) realloc (_hatab->vals,_maxi*sizeof(_VALTYPE)); \
      _hatab->maxkeys = _maxi;						\
      _oldcurkeys = _hatab->curkeys;                                    \
      _RESETF(_hatab,0);                                                \
      _hatab->curkeys = _oldcurkeys;                                    \
      for(_i=_hatab->curkeys; _i< _hatab->maxkeys;_i++) {               \
         _hatab->keys[_i] = 0;                                          \
         _hatab->vals[_i] = FORBIDDEN;                                  \
      }                                                                 \
      for(_i=0; _i<_hatab->curkeys;_i++)                                \
         if (_hatab->vals[_i] != FORBIDDEN)                             \
             g_hash_table_insert(_hatab->hash,			        \
		(gpointer)&(_hatab->keys[_i]),                    	\
		(gpointer)&(_hatab->vals[_i]));           	        \
    }									\
    _hatab->keys[_ind] = _key;						\
    _hatab->vals[_ind] = _val;						\
    g_hash_table_insert(_hatab->hash,					\
			(gpointer)&(_hatab->keys[_ind]),		\
			(gpointer)&(_hatab->vals[_ind]));		\
    _hatab->curkeys=(_ind+1>_hatab->curkeys?_ind+1:_hatab->curkeys);	\
  }

#define _GETPOINTER(_hatab,_keyx,_keyy,_N,_pval) {		     \
    _KEYTYPE _key = _KEYXY(_keyx,_keyy);			     \
    _pval=((_VALTYPE*)g_hash_table_lookup(_hatab->hash,(void*)&_key)); \
  }

#define _GETPDETAIL(_hatab,_keyx,_keyy,_p,_N,_pval) {		     \
    _pval=&(sdata->vals[p]);					     \
  }

#define _GETF(_hatab,_keyx,_keyy,_N,_val)	{		     \
    _KEYTYPE _key = _KEYXY(_keyx,_keyy);			     \
    _VALTYPE* _pval=((_VALTYPE*)g_hash_table_lookup(_hatab->hash,(void*)&_key)); \
    _val=((_pval == NULL)?NOTAVALUE:*(_pval));				\
  }



#ifndef _SUMCOLL_
#define _SUMCOLL_
#ifdef COUNT_COLL
typedef struct _MyGHashNode   MyGHashNode;
typedef struct _MyGHashTable  MyGHashTable;

struct _MyGHashNode {
  gpointer   key;
  gpointer   value;
  guint      key_hash;
};

struct _MyGHashTable {
  gint             size;
  gint             mod;
  guint            mask;
  gint             nnodes;
  gint             noccupied;  /* nnodes + tombstones */
  MyGHashNode      *nodes;
  GHashFunc        hash_func;
  GEqualFunc       key_equal_func;
  volatile gint    ref_count;
  int              version;
  GDestroyNotify   key_destroy_func;
  GDestroyNotify   value_destroy_func;
};

static inline guint lookup_coll (MyGHashTable    *hash_table,
		   gconstpointer  key, guint *nbcoll) {
  MyGHashNode *node;
  guint node_index;
  guint hash_value;
  guint step = 0;

  hash_value = (* hash_table->hash_func) (key);
  if (G_UNLIKELY (hash_value <= 1))
    hash_value = 2;

  node_index = hash_value % hash_table->mod;
  node = &hash_table->nodes [node_index];

  while (node->key_hash)
    {
      if (node->key_hash == hash_value)
        {
          if (hash_table->key_equal_func)
            {
              if (hash_table->key_equal_func (node->key, key))
                break;
            }
          else if (node->key == key)
            {
              break;
            }
        }
      step++;
#ifdef DISPLAY_COLL
      printf("ghash coll key %10u oldkey %10u index %10d step %4u\n",hash_value,node->key_hash,node_index,step);
#endif
      node_index += step;
      node_index &= hash_table->mask;
      node = &hash_table->nodes [node_index];
    }
  *nbcoll = step;
  return node_index;
}
  
static inline void sum_coll (GHashTable *hash) {
  MyGHashTable *hash_table;
  MyGHashNode *node;
  guint sumnbcoll;
  float avgcoll;
  guint sum, sumcoll, k, i, nbcoll;
  gint key;
  //  gint *pval;
  hash_table = (MyGHashTable*)hash;
  for (sum = 0, sumcoll = 0, k = 0; k < hash_table->size; k++) {
    node = &hash_table->nodes[k];
    if (node->key_hash > 1) {
      key  = *((gint*)node->key);
      i = lookup_coll(hash_table,&key,&nbcoll);
      //      pval = g_hash_table_lookup((GHashTable*)hash_table,&key);
      //      printf("ghash_lookup %2d %2d\n",key,*pval);
      //      printf("   my_lookup %2d %2d [%2d]\n",key,*((gint *)(hash_table->nodes[i]).value),nbcoll);
      sum++; sumcoll += nbcoll;
    }
  }
  sumnbcoll = sumcoll;
  avgcoll = (float)sumcoll/(float)sum;
  printf ("size %u\n",(unsigned)sizeof(guint));
  printf ("nb of collisions %d mean nb of entries probed while searching (ideal would be 1.): %f\n",sumnbcoll,1.+avgcoll);
}
#else

static inline void sum_coll (GHashTable *hash) {}

#endif
#endif

#define DEFKB ((long long)1<<10)

#define _GETKB(_ppset,_N) ((long long)(sizeof(hashtable)+_ppset->maxkeys*(sizeof(_KEYTYPE)+sizeof(_VALTYPE))) \
			/DEFKB)

#define _PGETKB(_ppset,_N) {					\
    printf("structure: %f, array of (key+value): %f, hashtable: ?\n",\
	   (1.*sizeof(hashtable))/DEFKB,			\
	   (1.*_ppset->maxkeys*(sizeof(_KEYTYPE)+sizeof(_VALTYPE)))/DEFKB); \
    sum_coll(_ppset->hash); \
    }


#endif
