#include <time.h>
#include <assert.h>
#include <ctype.h>
#include <malloc.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <dlfcn.h>
#include <unistd.h>
#include <sys/times.h>


long long vvms[10];
long long vrss[10];

/* Store the process's short name (as appears in ps) */
static char process_name[1024];

void get_mem_usage(long long *vms, long long *rss) {
  FILE *f;
  char procfn[32];
  char s[1024];

  *vms = 0;
  *rss = 0;

  snprintf(procfn, sizeof(procfn), "/proc/%d/status", getpid());
  procfn[sizeof(procfn)-1] = '\0';

  if ((f = fopen(procfn, "r")) == NULL) {
    fprintf(stderr, "get_mem_usage: Could not open %s: %s\n",
	    procfn, strerror(errno));
    return;
  }

  /* Search /proc/$pid/status for the lines we care about. */
  while (fgets(s, sizeof(s), f)) {
    /* Grab the process name, if we haven't already. */
    if (process_name[0] == '\0' && strncmp(s, "Name:", 5) == 0) {
      char *p;
      strncpy(process_name, s+6, sizeof(process_name));
      for (p = process_name; *p; p++)
	if (*p == '\r' || *p == '\n') {
	  *p = '\0';
	  break;
	}
    }
    else {
      //      printf("%s\n",s);
      if (strncmp(s, "VmRSS:", 6) == 0)
      {
	char usage_str[32];
	char *p;

	/* Strip leading whitespace. */
	for (p = s+8; isspace(*p); p++);
	strncpy(usage_str, p, sizeof(usage_str));

	/* Strip trailing whitespace */
	for (p = usage_str; *p && !isspace(*p); p++);
	*p = '\0';

	/* We only deal with kB at the moment. */
	assert(p[1] == 'k' && p[2] == 'B');
	*rss = atoll(usage_str);
	break;
      }
      if (strncmp(s, "VmSize:", 7) == 0)
      {
	char usage_str[32];
	char *p;

	/* Strip leading whitespace. */
	for (p = s+8; isspace(*p); p++);
	strncpy(usage_str, p, sizeof(usage_str));

	/* Strip trailing whitespace */
	for (p = usage_str; *p && !isspace(*p); p++);
	*p = '\0';

	/* We only deal with kB at the moment. */
	assert(p[1] == 'k' && p[2] == 'B');
	*vms = atoll(usage_str);
      }
    }
  }
  
  fclose(f);
}
