#include <math.h>
#include <assert.h>
#include "sstruct.h"

void ptree_init(ptrees **ppset, int shift, int N) {
  int         nblevel;
  int         coarse_size, cell_width, cell2d, grid2d;
  ptrees *pset;
  double      fnblevel;
  int         maxkeys = 1;
  fnblevel    = log2(((double)N));
  nblevel     = (int)fnblevel;
  cell_width  = 1<<shift;
  coarse_size = N>>shift;
  assert(cell_width * coarse_size == N);
  cell2d      = cell_width * cell_width;
  grid2d      = coarse_size * coarse_size;
  pset        = (ptrees *) malloc(sizeof(struct sptrees));
  *ppset      = pset;
  pset->coarse      = (_VALTYPE*) malloc (grid2d*sizeof(_VALTYPE));
  pset->indir       = (int*)      malloc (grid2d*sizeof(int));

  pset->ends        = (int*)      malloc((nblevel+2)*sizeof(int));
  pset->keys        = (_KEYTYPE*) malloc (maxkeys*sizeof(_KEYTYPE));
  pset->curkeys     = maxkeys;
  pset->maxkeys     = maxkeys;

  pset->stack       = (_VALTYPE*) malloc (cell2d*sizeof(_VALTYPE));
  pset->curcells    = 1;
  pset->maxcells    = 1;

//  pset->shift       = shift;
//  pset->shift2      = 2*shift;
  pset->coshift     = nblevel-shift;
  pset->cell_width  = cell_width;
  pset->coarse_size = coarse_size;
  pset->cell2d      = cell2d;
  pset->grid2d      = grid2d;
  pset->N           = N;
  pset->shift       = shift;
  pset->shift2      = 2*shift;
  pset->mask        = (1<<shift)-1;
  _RESETF(pset,0);
}

void ptree_destroy(ptrees **ppset) {
  ptrees *pset;
  pset = *ppset;
  free(pset->coarse);
  free(pset->indir);
  free(pset->stack);
  free(pset);
  *ppset = NULL;
}

