#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <malloc.h>
#include "sstruct.h"

#ifndef PTREE_H
#define PTREE_H


struct sptrees {
  _VALTYPE *coarse;
  int       coshift, cell_width, coarse_size;
  int       shift, shift2, mask;
  int       cell2d, grid2d, N;
  int      *indir;  
  _KEYTYPE *keys;      
  int      maxkeys, curkeys;
  int      *ends;
  _VALTYPE *stack;
  int       curcells, maxcells;
};

typedef struct sptrees ptrees;

void ptree_init(ptrees **ppset, int shift, int N);
void ptree_destroy(ptrees **ppset);

#define _INITF(_ppset,_shift,_size) ptree_init(_ppset,_shift,_size)

#define _DESTROY(_ppset) ptree_destroy(_ppset)

#define _RESETF(_pset,_size) {						\
    int _i;								\
    for (_i=0; _i<_pset->grid2d; _i++) {				\
      _pset->coarse[_i] = NOTAVALUE;					\
      _pset->indir [_i] = 0;						\
    }									\
    _pset->curcells = 1;						\
    _pset->curkeys  = 0;						\
    for (_i=0;_i<_pset->maxcells*_pset->cell2d; _i++) {			\
      _pset->stack[_i] = NOTAVALUE;					\
    }									\
  }

//      fprintf(stderr,"REALLOC KEYS %d\n",maxi);				
//	  fprintf(stderr,"REALLOC CELLS %d\n",maxi);			

#define _SETF_NEW(_pset,_keyx,_keyy,_N,_val) {				\
    int _cr = ((_keyy>>_pset->shift)<<(_pset->coshift))+(_keyx>>_pset->shift);	\
    int _lr = ((_keyy&_pset->mask)<<(_pset->shift))+(_keyx&_pset->mask);			\
    if (_pset->curkeys == _pset->maxkeys) {				\
      int maxi = _pset->maxkeys*1.2;					\
      maxi=(maxi>10?maxi:10);						\
      _pset->keys = (_KEYTYPE*) realloc (_pset->keys,maxi*sizeof(_KEYTYPE)); \
      _pset->maxkeys = maxi;						\
    }									\
    _pset->keys[_pset->curkeys++] = _KEYXY(_keyx,_keyy);		\
    if (_lr == 0) { _pset->coarse[_cr] = _val; }			\
    else {								\
      int _i, *indirect = &(_pset->indir[_cr]);				\
      if (*indirect == 0) {						\
	if (_pset->curcells == _pset->maxcells) {			\
	  int maxi = _pset->maxcells*1.2;				\
	  maxi=(maxi>10?maxi:10);					\
	  _pset->maxcells=maxi;						\
	  _pset->stack = (_VALTYPE*)					\
	    realloc(_pset->stack,					\
		    _pset->maxcells*_pset->cell2d*sizeof(_VALTYPE));	\
	  for (_i=1+(_pset->curcells)*_pset->cell2d;			\
	       _i<_pset->maxcells*_pset->cell2d; _i++) {		\
	    _pset->stack[_i] = NOTAVALUE;				\
	  }								\
	}								\
	*indirect=_pset->curcells;					\
	_pset->curcells++;						\
      }									\
      _pset->stack[(*indirect)*_pset->cell2d+_lr]=_val;			\
    }									\
  }

//    if (_keyx==0) {							
//      printf("new coord %d %d ind %d %d %d %f\n",_keyx,_keyy,*indirect,_cr,_lr,_val); 
//      printf("new info %d %d %d\n",(_keyy>>_pset->shift),_pset->coarse_size,(_keyx>>_pset->shift)); 
//    }									

#define _GETPOINTER(_pset,_keyx,_keyy,_N,_pval) {			\
    int _cr = ((_keyy>>_pset->shift)<<(_pset->coshift))+(_keyx>>_pset->shift);	\
    int _lr = ((_keyy&_pset->mask)<<(_pset->shift))+(_keyx&_pset->mask);			\
    if (_lr == 0) _pval = &(_pset->coarse[_cr]);			\
    else {								\
      _pval = &(_pset->stack[((_pset->indir[_cr])<<(_pset->shift2))+_lr]);	\
    }									\
    if ((*_pval) == NOTAVALUE) {					\
      _pval = NULL;							\
    }									\
  }
#define _GETPDETAIL(_pset,_keyx,_keyy,_p,_N,_pval) \
  _GETPOINTER(_pset,_keyx,_keyy,_N,_pval)


#define _GETF(_pset,_keyx,_keyy,_N,_val)	{			\
    int _cr = ((_keyy>>_pset->shift)<<(_pset->coshift))+(_keyx>>_pset->shift);	\
    int _lr = ((_keyy&_pset->mask)<<(_pset->shift))+(_keyx&_pset->mask);			\
    if (_lr == 0) _val = (_pset->coarse[_cr]);				\
    else {								\
      _val = (_pset->stack[((_pset->indir[_cr])<<(_pset->shift2))+_lr]);	\
    }									\
  }

#define _GETCOARSE_PT(_pset,_keyx,_keyy,_pval) {			\
    int _cr = ((_keyy>>_pset->shift)<<(_pset->coshift))+(_keyx>>_pset->shift);	\
    int _lr = ((_keyy&_pset->mask)<<(_pset->shift))+(_keyx&_pset->mask);			\
    assert (_lr == 0);							\
    _pval = &(_pset->coarse[_cr]);					\
    if ((*_pval) == NOTAVALUE) {					\
      _pval = NULL;							\
    }									\
  }

#define _CELLEXIST(_pset,_keyx,_keyy)					\
  (_pset->indir[((_keyy>>_pset->shift)<<(_pset->coshift))+		\
		(_keyx>>_pset->shift)] != 0)				\
 
//#define DEFKB ((long long)1000)
#define DEFKB ((long long)1<<10)

#define _GETKB(_ppset,_N) ((long long)(sizeof(ptrees)+_ppset->maxcells*_ppset->cell2d*sizeof(_VALTYPE)+ \
				    _ppset->maxkeys*sizeof(_KEYTYPE)+_ppset->grid2d*(sizeof(int)+sizeof(_VALTYPE))) \
			/DEFKB)

#define _PGETKB(_ppset,_N) {						\
    printf("structure: %f, stack: %f, key list: %f, coarse grid: %f\n",	\
	   1.*sizeof(ptrees)/DEFKB,			\
	   1.*_ppset->maxcells*_ppset->cell2d*sizeof(_VALTYPE)/DEFKB, \
	   1.*_ppset->maxkeys*sizeof(_KEYTYPE)/DEFKB,	\
	   1.*_ppset->grid2d*(sizeof(_KEYTYPE)+sizeof(_VALTYPE))/DEFKB);}


#endif
