#include <stdlib.h>
#include <assert.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include "wav.h"
#include "meminfo.h"
#include "clock.h"

/* Sizes of the predict and update filters */
int wav_nu=0, wav_np = 0;
/* Offset for the predict and update filters */
int wav_offu=0, wav_offp = 0;
/* Arrays that store wavelet filters coefficients */ 
double *wav_filterPred = NULL;
double *wav_filterUpd  = NULL;


int mylog2(int input) {
  return((int)round(log2(input)));
}

int inline wav_ilevel(int ix, int iy, int nblevel) {
  int j = ((_KEYTYPE)0x1)<<nblevel, lev;
  for(j |= (ix|iy), lev = 0; 
      ((j&(_KEYTYPE)0x1) == 0); j>>=1, lev++);
  return(lev);
}


void wav_init(int type) {
  if (wav_filterUpd != NULL) free(wav_filterUpd);
  if (wav_filterPred != NULL) free(wav_filterPred);
  switch (type) {
  case 0: //HAT : DESLAURIERS-DUBUC FILTER N=2
    wav_np = 2;
    wav_offp = 0;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = 1./2.;
    wav_filterPred[1] = 1./2.;
    wav_nu = 0;
    break;

  case 1: // DESLAURIERS-DUBUC FILTER N=4
    wav_np = 4;
    wav_offp = -1;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = -1./16.;
    wav_filterPred[1] =  9./16.;
    wav_filterPred[2] =  9./16.;
    wav_filterPred[3] = -1./16.;
    wav_nu = 0;
    break;

  case 2: // DESLAURIERS-DUBUC FILTER N=6
    wav_np = 6;
    wav_offp = -2;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] =   3./256.;
    wav_filterPred[1] = -25./256.;
    wav_filterPred[2] = 150./256.;
    wav_filterPred[3] = 150./256.;
    wav_filterPred[4] = -25./256.;
    wav_filterPred[5] =   3./256.;
    wav_nu = 0;
    break;

  case 3: //COIFFLET
    wav_np = 6;
    wav_offp = -2;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = -0.0727326;
    wav_filterPred[1] = 0.337898;
    wav_filterPred[2] = 0.852572;
    wav_filterPred[3] = 0.384865;
    wav_filterPred[4] = -0.0727326;
    wav_filterPred[5] = -0.0156557;
    wav_nu = 0;
    wav_offu = 0;
    break;

  case 10: //INTER_N2_NTILDE2
    wav_np = 2;
    wav_offp = 0;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = 1./2.;
    wav_filterPred[1] = 1./2.;
    wav_nu = 2;
    wav_offu = 0;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] = 1./4.;
    wav_filterUpd[1] = 1./4.;
    break;

  case 11: //INTER_N4_NTILDE2
    wav_np = 4;
    wav_offp = -1;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = -1./16.;
    wav_filterPred[1] =  9./16.;
    wav_filterPred[2] =  9./16.;
    wav_filterPred[3] = -1./16.;
    wav_nu = 2;
    wav_offu = 0;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] = 1./4.;
    wav_filterUpd[1] = 1./4.;
    break;

  case 12: 
    wav_np = 4;
    wav_offp = -1;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = -1./16.;
    wav_filterPred[1] =  9./16.;
    wav_filterPred[2] =  9./16.;
    wav_filterPred[3] = -1./16.;
    wav_nu = 4;
    wav_offu = -1;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] = -1./16.;
    wav_filterUpd[1] =  5./16.;
    wav_filterUpd[2] =  5./16.;
    wav_filterUpd[3] = -1./16.;
    break;
    
  case 13: 
    wav_np = 2;
    wav_offp = 0;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = 1./2.;
    wav_filterPred[1] = 1./2.;
    wav_nu = 4;
    wav_offu = -1;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] = -1./16.;
    wav_filterUpd[1] =  5./16.;
    wav_filterUpd[2] =  5./16.;
    wav_filterUpd[3] = -1./16.;
    break;
    
  case 14: //INTER_N4_NTILDE6:
    wav_np = 4;
    wav_offp = -1;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = -1./16.;
    wav_filterPred[1] =  9./16.;
    wav_filterPred[2] =  9./16.;
    wav_filterPred[3] = -1./16.;
    wav_nu = 6;
    wav_offu = -2;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] =    9./1024.;
    wav_filterUpd[1] = - 59./1024.;
    wav_filterUpd[2] =  306./1024.;
    wav_filterUpd[3] =  306./1024.;
    wav_filterUpd[4] = - 59./1024.;
    wav_filterUpd[5] =    9./1024.;
    break;

  case 15: //HAAR
    wav_np = 1;
    wav_offp = 0;
    wav_filterPred = (double*)malloc(wav_np*sizeof(double));
    wav_filterPred[0] = 1.;
    wav_nu = 1;
    wav_offu = 1;
    wav_filterUpd = (double*)malloc(wav_nu*sizeof(double));
    wav_filterUpd[0] = 1./2.;
    break;

  default:
    printf("Wavelet type undefined\n");exit(0);
  }
}

/*
  Wavelet Transform.
  
  The terms "predict" and "update" comes from the article
  'Building Your Own Wavelets at Home' from Sweldens

  - Transform

  In the wavelet transform, calculate the difference between 
  an odd element and its approximation-interpolation with 
  even coarser points. 
   
  - Inverse Transform

  Reverse the transform predict step by adding the coarse approximation
  (even element) to the difference-detail (odd element).

  Parameters
  - fwav        : the input data on which the forward or inverse
                  transform is calculated.
  - dir(ection) : (1) forward or (-1) inverse transform

*/



/*
  Wavelet transform.
  In-place wavelet algorithm is used here.
  A dense 2D array (fwav) of N*N elements is
  passed to the 2D transform function. 
  The result of the wavelet scheme is calculated
  in-place (without any array temporaries) in the 
  array 'fwav'.
*/
void wav_ftransform( _VALTYPE *fwav, int N ) {
  int itc;
  for (itc = 2; itc <= N; itc = itc * 2) {
    wav_predict_x ( fwav, itc, N, 1 );
    wav_update_x  ( fwav, itc, N, 1 );
    wav_predict_y ( fwav, itc, N, 1 );
    wav_update_y  ( fwav, itc, N, 1 );
  } 
} 

/*
  Wavelet inverse transform.
  Same comments as  wav_ftransform function
*/
void wav_itransform( _VALTYPE *fwav, int N ) {
  int itc;
  for (itc = N; itc >= 2; itc = itc / 2) {
    wav_update_y  ( fwav, itc, N, -1 );
    wav_predict_y ( fwav, itc, N, -1 );
    wav_update_x  ( fwav, itc, N, -1 );
    wav_predict_x ( fwav, itc, N, -1 );
  }
}

/* Predict function in x-direction 
   fwav: input/output array
   itc: level+1 which is considered
   N: square root of the 2D domain size
   dir: (1=forward) or (-1=inverse) transform
   wav_filterPred: predict filter
   wav_np: size of the predict filter
 */
inline void wav_predict_x(_VALTYPE* fwav, int itc, 
			  int N, int dir ) {
  register int ix, iy, icoarse;
  register const int itf = (itc>>1); /* itc/2 */
  register int k, Nplus = N*N;
  register double predictVal;
  /* Scan all iy at level 'itf' */
  for (iy = 0; iy < N; iy += itf) {
    /* Apply predict operator in x-direction */
    for (ix = itf; ix < N; ix += itc) {
      /* At index ix is the detail point we are 
	 looking at, compute predict value */
      predictVal = 0.;
      for (k = 0; k < wav_np; k++) {
	icoarse = (Nplus + ix + 
		   (k + wav_offp)*itc - itf)%N;
	predictVal += wav_filterPred[k] * 
	  fwav[_XY(icoarse,iy,N)];
      }
      /* Add or retrieve the predict value to 
	 the 'odd' points whether the forward 
	 or inverse transform is applied */
      fwav[_XY(ix,iy,N)] += - dir * predictVal;
    } 
  }
}

/* Update function in x-direction 
   fwav: input/output array
   itc: level+1 which is considered
   N: square root of the 2D domain size
   dir: (1=forward) or (-1=inverse) transform
   wav_filterPred: predict filter
   wav_np: size of the predict filter
*/
inline void wav_update_x(_VALTYPE* fwav, int itc,
			 int N, int dir ) {
  register int ix, iy, idetail;
  register const int itf = (itc>>1); /* itc/2 */
  register int k, Nplus = N*N;
  register double updateVal;
  /* Scan all iy at level 'itf' */
  for (iy = 0; iy < N; iy += itf) {
    /* Apply update operator in x-direction */
    for (ix = 0; ix < N; ix += itc) {
      /* At index ix is the coarse point we
	 are looking at, compute udapte value */
      updateVal = 0.;
      for (k = 0; k < wav_nu; k++) {
	idetail = (Nplus + ix + 
		   (k + wav_offu)*itc - itf)%N;
	updateVal += wav_filterUpd[k] * 
	  fwav[_XY(idetail,iy,N)];
      }
      /* Add or retrieve the update value to 
	 the 'even'points whether the forward 
	 or inverse transform is applied */
      fwav[_XY(ix,iy,N)] += dir * updateVal;
    } 
  }
}

/* Predict function in y-direction 
   fwav: input/output array
   itc: level+1 which is considered
   N: square root of the 2D domain size
   dir: (1=forward) or (-1=inverse) transform
   wav_filterPred: predict filter
   wav_np: size of the predict filter
*/
inline void wav_predict_y(_VALTYPE* fwav, int itc, 
			  int N, int dir ) {
  register int ix, iy, icoarse;
  register const int itf = (itc>>1); /* itc/2 */
  register int k, Nplus = N*N;
  register double predictVal;

  /* Apply predict operator in y-direction */
  for (iy = itf; iy < N; iy += itc) {
    /* Scan all ix at level 'itf' */
    for (ix = 0; ix < N; ix += itf) {
      /* At index iy is the detail point we are 
	 looking at, compute predict value */
      predictVal = 0.;
      for (k = 0; k < wav_np; k++) {
	icoarse = (Nplus + iy +
		   (k + wav_offp)*itc - itf)%N;
	predictVal += wav_filterPred[k] * 
	  fwav[_XY(ix,icoarse,N)];
      }
      /* Add or retrieve the predict value to 
	 the 'odd' points whether the forward 
	 or inverse transform is applied */
      fwav[_XY(ix,iy,N)] += - dir * predictVal;
    } 
  }
}

/* Update function in y-direction 
   fwav: input/output array
   itc: level+1 which is considered
   N: square root of the 2D domain size
   dir: (1=forward) or (-1=inverse) transform
   wav_filterUpd: update filter
   wav_nu: size of the update filter
*/
inline void wav_update_y(_VALTYPE* fwav, int itc, 
			 int N, int dir ) {
  register int ix, iy, idetail;
  register const int itf = (itc>>1); /* itc/2 */
  register int k, Nplus = N*N;
  register double updateVal;
  
  /* Apply predict operator in y-direction */
  for (iy = 0; iy < N; iy += itc) {
    /* Scan all ix at level 'itf' */
    for (ix = 0; ix < N; ix += itf) {
      /* At index iy is the coarse point we
	 are looking at, compute update value */
      updateVal = 0.;
      for (k = 0; k < wav_nu; k++) {
	idetail = (Nplus + iy + 
		   (k + wav_offu)*itc - itf)%N;
	updateVal += wav_filterUpd[k] * 
	  fwav[_XY(ix,idetail,N)];
      }
      /* Add or retrieve the update value to 
	 the 'even' points whether the forward 
	 or inverse transform is applied */
      fwav[_XY(ix,iy,N)] += dir * updateVal;
    } 
  }
}


/*
  Thresholding of coefficients
*/
void wav_thresholding( _VALTYPE *fwav, int N, double normcoef , 
		       double maxthreshold, double zerovalue, char *grid, 
		       int    *advised_shift) {
  int itc, itf, itv, nnz_tot, ix, iy;
  double *threshold_level;
  double nbpts_level;
  int Nplus = N*N;
  int *nnz_level;
  int nblevel, level, icoarse, idetail, k;
  nblevel = mylog2(N);
  threshold_level = (double*) malloc (nblevel*sizeof(double));
  nnz_level = (int*) malloc (nblevel*sizeof(int));
  for (level = 0; level < nblevel; level++) {
    itc = pow(2,nblevel-level);
    threshold_level[level] = maxthreshold * pow(normcoef, level);
    nnz_level[level] = 0;
  }
  nnz_tot = 1; /* counting coarse coefficient root of the wavelet tree */

  level = nblevel-1;
  for (iy = 0; iy < N; iy ++) 
    for (ix = 0; ix < N; ix ++)  
      grid[_XY(ix,iy,N)] = 0;

  /* Look at finest details */
  for (iy = 0; iy < N; iy += 1) {
    if ((iy & 1) == 0) {
      for (ix = 1; ix < N; ix += 2) {
	if (fabs(fwav[_XY(ix,iy,N)]) >= threshold_level[level]) {
	  grid[_XY(ix,iy,N)] = 1;
	  nnz_level[level]++;      
	}
      }
    } else {
      for (ix = 0; ix < N; ix += 1) {
	if (fabs(fwav[_XY(ix,iy,N)]) >= threshold_level[level]) {
	  grid[_XY(ix,iy,N)] = 1;
	  nnz_level[level]++;      
	}
      }
    }
  }

  /* Look at details of coarser levels */
  for (level = nblevel-2, itv = 1, itf = 2, itc = 4; itc <= N; 
       itc *= 2, itf *= 2, itv *=2, level--) {
    for (iy = 0; iy < N; iy +=itf) {
      if ((iy & itf) == 0) {
	for (ix = itf; ix < N; ix +=itc) {
	  if (fabs(fwav[_XY(ix,iy,N)]) >= threshold_level[level]) {
	    grid[_XY(ix,iy,N)] = 1;
	    nnz_level[level]++;      
	  }
	}
      } else {
	for (ix = 0; ix < N; ix +=itf) {
	  if (fabs(fwav[_XY(ix,iy,N)]) >= threshold_level[level]) {
	    grid[_XY(ix,iy,N)] = 1;
	    nnz_level[level]++;      
	  }
	}
      } 
    }
  }

  for (level = 0; level < nblevel; level++) {
    nbpts_level = .75*pow(2,2*(level+1));
    WAVDEBUG(printf("level %3d N %d threshold %.3e nnz %20d ratio %.4e\n",
		    level,N,threshold_level[level],nnz_level[level],nnz_level[level]/nbpts_level));
    nnz_tot += nnz_level[level]; 
 }
  printf("Number of non-zero coefficients :"				\
	 "%d over %d (%.3f percents)\n",nnz_tot,N*N,(100.*nnz_tot)/(N*N));

  /* Make a well builded tree */
  /* -> loop on levels from coarse to fine */
  /* -> apply an algorithm very close to forward transform */
  for (itf = 1 ; itf < N ; itf <<= 1){
    itc = 2*itf;

    /* Predict x */
    for (iy = 0; iy < N; iy += itf) {
      for (ix = itf; ix < N; ix += itc) {
	if (grid[_XY(ix,iy,N)]) {
	  for (k = 0; k < wav_np; k++) {
	    icoarse = (Nplus+ ix + (k + wav_offp)*itc - itf)%N;
	    grid[_XY(icoarse,iy,N)] = 1;
	  }
	}
      }
    }
    /* Update x */
    for (iy = 0; iy < N; iy += itf) {
      for (ix = 0; ix < N; ix += itc) {
	if (grid[_XY(ix,iy,N)]) {
	  for (k = 0; k < wav_nu; k++) {
	    idetail = (Nplus + ix + (k + wav_offu)*itc - itf)%N;
	    grid[_XY(idetail,iy,N)] = 1;
	  }
	}
      }
    }
    /* Predict y */
    for (iy = itf; iy < N; iy += itc) {
      for (ix = 0; ix < N; ix += itf) {
	if (grid[_XY(ix,iy,N)]) {
	  for (k = 0; k < wav_np; k++) {
	    icoarse = (Nplus + iy + (k + wav_offp)*itc - itf)%N;
	    grid[_XY(ix,icoarse,N)] = 1;
	  }
	}
      }
    }
    /* Update y */
    for (iy = 0; iy < N; iy += itc) {
      for (ix = 0; ix < N; ix += itf) {
	if (grid[_XY(ix,iy,N)]) {
	  for (k = 0; k < wav_nu; k++) {
	    idetail = (Nplus + iy + (k + wav_offu)*itc - itf)%N;
	    grid[_XY(ix,idetail,N)] = 1;
	  }
	}
      }
    }
  }

  /* In the data 'grid', the value 1 has been set
     at locations where details are non-zero or
     where details are required because of the wavelet
     tree structure */
  nnz_tot = 1;
  for (iy = 0; iy < N; iy ++) 
    for (ix = 0; ix < N; ix ++) {
      if (grid[_XY(ix,iy,N)] == 0) {
	fwav[_XY(ix,iy,N)] = zerovalue;
      } else {
	nnz_tot ++;
      }
    }


  /* Count non-zero coefficients at each level */
  for (level = nblevel-1, itf = 1, itc = 2; itc <= N; 
       itc *= 2, itf *= 2, level--) {
    nnz_level[level] = 0;
    for (iy = 0; iy < N; iy +=itf) {
      if ((iy & itf) == 0) {
	for (ix = itf; ix < N; ix +=itc) {
	  if (grid[_XY(ix,iy,N)]) {
	    nnz_level[level]++;      
	  }
	}
      } else {
	for (ix = 0; ix < N; ix +=itf) {
	  if (grid[_XY(ix,iy,N)]) {
	    nnz_level[level]++;      
	  }
	}
      }
    }
  }

  nnz_tot=1;
  for (level = 0; level < nblevel; level++) {
    nbpts_level = nnz_level[level]/(.75*pow(2,2*(level+1)));
    WAVDEBUG(printf("level %3d N %d nnz %20d ratio %.4e\n",
		    level,N,nnz_level[level],nbpts_level));
    nnz_tot += nnz_level[level];
  }
  printf("Number of non-zero coefficients :"				\
	 "%d over %d (%.3f percents)\n",nnz_tot,N*N,(100.*nnz_tot)/(N*N));
  /* Try to find a good 'shift'. This parameter controls
     the ptree data structure initialization (see ptree.c) */
  *advised_shift = 2;
  nbpts_level    = 0.;
  level = nblevel - 1 - *advised_shift;
  for (; (level > 0) && (nbpts_level < .1); level--) {
    nbpts_level = nnz_level[level]/(.75*pow(2,2*(level+1)));
    if (nbpts_level < .07) (*advised_shift)++;
  }
  printf("Advised shift %d\n",*advised_shift);
  free(nnz_level);
  free(threshold_level);
}

/*
  Copy an array from fold to fnew
*/
void wav_copyfunc_aa( _VALTYPE *fold, _VALTYPE *fnew, int N ) {
  register int ix, iy;
  for (iy = 0; iy < N; iy ++) 
    for (ix = 0; ix < N; ix ++)  
      fnew[_XY(ix,iy,N)] = fold[_XY(ix,iy,N)];
}

/*
  Copy an array from fold to fnew selecting point
  with (grid[ix,iy] == 1)
*/
void wav_copyfunc_aa_select( _VALTYPE *fold, _VALTYPE *fnew, int N, char *grid, double value ) {
  register int ix, iy;
  for (iy = 0; iy < N; iy ++) 
    for (ix = 0; ix < N; ix ++)  
      if (grid[_XY(ix,iy,N)]) 
	fnew[_XY(ix,iy,N)] = fold[_XY(ix,iy,N)];
      else
	fnew[_XY(ix,iy,N)] = value;
}
/* ============================= OPTIMIZED ALGORITHMS  ========================= */

/*
  Adaptive wavelet inverse transform.
*/
void wav_adapt_itransform( FTYPE *sold, int N) {
  const int dir = -1;
  int itc;
  /* loop on levels from coarse to fine */
  for (itc = N ; itc >= 2 ; itc >>= 1){
    wav_adapt_predict_y(sold, itc, N, dir);
    wav_adapt_predict_x(sold, itc, N, dir);
  }
}




inline 
void 
wav_inside_pred_y(FTYPE *sdata, int ix, 
		  int iy, int itf, int itc, 
		  int N, int dir, int Nplus, 
		  _VALTYPE *pdetail) {
  register int k, icoarse;
  double   predictVal, readval;
  /* iy is the detail index we are looking at */
  predictVal = 0;
  /* Apply predict operator in y-direction */
  for (k = 0; k < wav_np; k++) {
    icoarse = (Nplus + iy + 
	       (k + wav_offp)*itc - itf)%N;
    _GETF(sdata,ix,icoarse,N,readval);
    predictVal += wav_filterPred[k] * readval;
  }
  *pdetail += - dir * predictVal;
}


//    printf("level 0: itc=%d [%d:%d]\n",itc,pstart,pend);
//    printf("level %d: itc=%d [%d:%d]\n",ilevel,itc,
//	   pstart,pend);

#if defined(_HASH) || defined(_PTREE)

void wav_adapt_predict_y(FTYPE *sdata, int itc, 
			 int N, int dir) {
  _KEYTYPE ix, iy, idetail;
  _VALTYPE *pdetail;
  int      itf = itc / 2, Nplus = N*N;
  int      p, pstart, pend, ilevel;
  ilevel = mylog2(itf);
  if (ilevel == 0) { 
    pstart = 0; pend = sdata->ends[0]; 
  } else { 
    pstart = sdata->ends[ilevel-1]; 
    pend = sdata->ends[ilevel]; 
  }
  /*  Iterate over non-zero wavelet coefficients 
      at level parameterized by itf */
  for (p = pstart; p < pend; p++) {
    /* Get 'idetail': a key from the list. */
    idetail    = sdata->keys[p];
    /* Extract indices (ix,iy) from 'idetail' */
    _INVKEYXY(idetail,ix,iy);
    /*  With 'iy', determine wether the point 
	(ix,iy) is a detail at level 'itf' */
    if ((iy&itf) != 0) {
      /* Get pointer to memory area 'pdetail'
	 where detail val. will be stored  */
      _GETPDETAIL(sdata,ix,iy,p,N,pdetail);
      /* Compute the prediction and store
	 difference at 'pdetail' */
      wav_inside_pred_y(sdata, ix, iy, itf, itc,
			N, dir, Nplus, pdetail);
    }
  }
}

#else /* _HASH or _PTREE */

void wav_adapt_predict_y(FTYPE *sdata, int itc, 
			 int N, int dir) {
  _KEYTYPE ix, iy;
  int      itf = itc / 2;
  int      Nplus = N*N;
  _VALTYPE *pdetail;

  for (iy = itf; iy < N; iy += itc) 
    for (ix = 0; ix < N; ix += itf) {
      _GETPOINTER(sdata,ix,iy,N,pdetail);
      if (pdetail != NULL) {
	wav_inside_pred_y(sdata, ix, iy, itf, itc,
			     N, dir, Nplus, pdetail);
      }
    }
}

#endif /* _HASH or _PTREE */

inline 
void wav_inside_pred_x(FTYPE *sdata, int ix, int iy, 
			  int itf, int itc, int N, int dir, 
			  int Nplus, _VALTYPE *pdetail) {
  register int k, icoarse;
  double   predictVal, readval;
  predictVal = 0;
#ifdef STOFFSET
  for (k = STOFFSET; k < wav_np-STOFFSET; k++) {
#else
  for (k = 0; k < wav_np; k++) {
#endif
    icoarse = (Nplus + ix + (k + wav_offp)*itc - itf)%N;
    _GETF(sdata,icoarse,iy,N,readval);
    predictVal += wav_filterPred[k] * readval;
  }
  *pdetail += - dir * predictVal;
}

void wav_adapt_predict_x(FTYPE *sdata, int itc, 
			 int N, int dir) {
  _KEYTYPE ix, iy;
  int      itf = itc / 2;
  int      Nplus = N*N;
  _VALTYPE   *pdetail;

  /* No update step implemented in the adaptive 
     version of the inverse wavelet transform */
  assert(wav_nu == 0);

#if defined(_HASH) || defined(_PTREE)
  int      p, pstart, pend, ilevel;
  _KEYTYPE idetail;
  ilevel = mylog2(itf);
  if (ilevel == 0) pstart = 0;
  else             pstart = sdata->ends[ilevel-1];
  pend   = sdata->ends[ilevel];
  for (p=pstart; p<pend; p++) {
    idetail    = sdata->keys[p];
    _INVKEYXY(idetail,ix,iy);
    if ((ix&itf) != 0) {
      _GETPDETAIL(sdata,ix,iy,p,N,pdetail);
      wav_inside_pred_x(sdata, ix, iy, itf, itc,
			   N, dir, Nplus, pdetail);
    }
  }
#else
  for (iy = 0; iy < N; iy += itf) 
    for (ix = itf; ix < N; ix += itc) {
      _GETPOINTER(sdata,ix,iy,N,pdetail);
      /* Select only non zero elements at this
	 level identified by itf */
      if (pdetail != NULL) {
	wav_inside_pred_x(sdata, ix, iy, itf, itc,
			   N, dir, Nplus, pdetail);
      }
    }
#endif /* _HASH or _PTREE */
}


void wav_reorder_grid( FTYPE *sold, int N ) {
#if defined(_HASH) || defined(_PTREE)
  int p,i;
  _KEYTYPE key,kx,ky;
  int nblevel,ilevel,curkeys;
  int *hole;
  _KEYTYPE *oldkeys;
#ifdef _HASH
  _VALTYPE val;
  _VALTYPE *oldvals;
  oldvals = (_VALTYPE*) malloc (sold->curkeys*sizeof(_VALTYPE));
#endif

  curkeys = sold->curkeys;
  nblevel = mylog2(N);
  //  printf("nblevel %d %d\n",N,nblevel);
  hole    = (int*) malloc((2+nblevel)*sizeof(int));
  oldkeys = (_KEYTYPE*) malloc (sold->curkeys*sizeof(_KEYTYPE));
  for (ilevel=0; ilevel<2+nblevel; ilevel++) {
    hole[ilevel] = 0;
  }
  for (p=0; p<curkeys; p++) {
    key     = sold->keys[p];
    oldkeys[p] = key;
#ifdef _HASH
    oldvals[p] = sold->vals[p];
#endif
    _INVKEYXY(key,kx,ky);
    ilevel = wav_ilevel(kx,ky,nblevel);
    hole[ilevel]++;
  }
  sold->ends[0] = hole[0];
  for (ilevel=1; ilevel<1+nblevel; ilevel++) {
    sold->ends[ilevel] = hole[ilevel] + sold->ends[ilevel-1];
  }
  hole[0]=0;
  for (ilevel=1; ilevel<1+nblevel; ilevel++) {
    hole[ilevel] = sold->ends[ilevel-1];
  }
  
#ifdef _HASH
  _RESETF(sold,N);
  for (p=0; p<curkeys; p++) {
    key = oldkeys[p];
    val = oldvals[p];
    _INVKEYXY(key,kx,ky);
    ilevel = wav_ilevel(kx,ky,nblevel);
    i = hole[ilevel];
    _SETF_MODIFY_K(sold,key,val,i);
    hole[ilevel]++;
  }
  free(oldvals);
  free(oldkeys);
#else
  for (p=0; p<curkeys; p++) {
    key = oldkeys[p];
    _INVKEYXY(key,kx,ky);
    ilevel = wav_ilevel(kx,ky,nblevel);
    i = hole[ilevel];
    sold->keys[i] = key;
    hole[ilevel]++;
  }
  free(oldkeys);
#endif
  for (ilevel=0; ilevel<1+nblevel; ilevel++) 
    printf("reorder %d hole %d ilev %d\n",
	   ilevel,hole[ilevel],sold->ends[ilevel]);
  printf("curkeys %d == %d\n",sold->curkeys,curkeys);
  free(hole);
#endif
    
}

/*
  Copy an array from fold to snew
*/
void wav_copyfunc_as( _VALTYPE *fold, FTYPE *snew, int N) {
  register _KEYTYPE ix, iy;//, kk, kx, ky;
  register double val;
  get_mem_usage(&(vvms[1]), &(vrss[1]));
   _RESETF(snew,N);
  for (iy = 0; iy < N; iy ++) {
    for (ix = 0; ix < N; ix ++) {
      val = fold[_XY(ix,iy,N)];
      if (val != NOTAVALUE){
	_SETF_NEW(snew,ix,iy,N,val);
      }
    }
  }
  get_mem_usage(&(vvms[2]), &(vrss[2]));

#if defined(_HASH) || defined(_PTREE)
  GETTIME(wav_reorder_grid(snew,N),"<<reorder>>");
#endif
  get_mem_usage(&(vvms[3]), &(vrss[3]));

//  { 
//    int ilevel, nblevel = mylog2(N);
//    for (ilevel=0; ilevel<1+nblevel; ilevel++) 
//      printf("copy %d  = %d\n",ilevel,
//	     snew->ends[ilevel]);
//  }
#ifdef _PTREE
//  {
//    int kk;
//    for (kk = 0; kk < snew->curkeys; kk ++) {
//      _INVKEYXY(snew->keys[kk],ix,iy);
//      printf("KEYS %d %d: %d %d\n",snew->keys[kk],kk,ix,iy);
//    }
//  }
#endif
}

/*
  Copy an array from fold to snew
*/
void wav_copyfunc_ss( FTYPE *sold, FTYPE *snew, int N) {
  _VALTYPE val;
#ifdef _HASH
  int ilevel, nblevel, p;
  _KEYTYPE key;
  _RESETF(snew,N);
  for (p=0; p<sold->curkeys; p++) {
    key = sold->keys[p];
    val = sold->vals[p];
    _SETF_MODIFY_K(snew,key,val,p);
  }
  nblevel = mylog2(N);
  for (ilevel=0; ilevel<nblevel+1; ilevel++) {
    snew->ends[ilevel] = sold->ends[ilevel];
  }
#else
  int ix,iy;
  _RESETF(snew,N);
  for (iy = 0; iy < N; iy ++) {
    for (ix = 0; ix < N; ix ++) {
      _GETF(sold,ix,iy,N,val);
      if (val != NOTAVALUE){
	_SETF_NEW(snew,ix,iy,N,val);
      }
    }
  }
#endif
}


/*
  Adaptive wavelet forward transform.
*/
void wav_adapt_ftransform( FTYPE *sold, int N) {
  const int dir = 1;
  int itc;
  /* No update step implemented in the adaptive 
     version of the inverse wavelet transform */
  assert(wav_nu == 0);

  /* loop on levels from fine to coarse*/
  for (itc = 2 ; itc <= N ; itc *= 2){
    wav_adapt_predict_x(sold, itc, N, dir);
    wav_adapt_predict_y(sold, itc, N, dir);
  }
}

void wav_getkb( FTYPE *sold, int N) {
  int nbmb = _GETKB(sold,N);
  _PGETKB(sold,N);
  printf("Size of structure %d KB\n",nbmb);
}

void wav_flushcache(int mbsize) {
  double *memp;
  int     bsize = (1<<20)*mbsize;
  int     nbelt = bsize/sizeof(double);
  memp = (double *) malloc(nbelt*sizeof(double));
  memset(memp,'\1',bsize),
  free(memp);
}
