#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include "clock.h"
#include "meminfo.h"
#include "wav.h"
#ifndef DRATE
#define DRATE 1000000000
#endif
#ifndef TSIZE
#define TSIZE 1024
#endif
#ifndef MODEX
#define MODEX 8
#endif
#ifndef COEFFEXP
#define COEFFEXP -0.2
#endif



/* Write into a file the contents of the array "data" */
/* The name of the file is "fname", the "size" specifies the array size */
void print2d(_VALTYPE *data, int size, double dx, char *fname) {
  FILE*  fd;
  register int ix, iy;
  register double val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",fname);
    exit(-1);
  }
  for (iy = 0; iy < size; iy++) {
    for (ix = 0; ix < size; ix++) {
      val = data[_XY(ix,iy,size)];
      fprintf(fd,"%f %f %.3e\n",dx*ix,dx*iy,val);
    }
    fprintf(fd,"\n");
  }
  fclose(fd);
}

void sparse_print2d(FTYPE *sdata, int size, 
 		    double dx, char *fname) {
  FILE*    fd;
  _KEYTYPE ix, iy;
  _VALTYPE val;
  if ((fd=fopen(fname,"w")) == NULL) {
    printf("file %s could not be opened\n",
 	   fname);
    exit(-1);
  } else {
    for (iy = 0; iy < size; iy++) {
      for (ix = 0; ix < size; ix++) {
	_GETF(sdata,ix,iy,size,val);
	if (val != NOTAVALUE)
	  fprintf(fd,"%f %f %.3e\n",dx*ix,dx*iy,val);
      }
    }
  }
  fclose(fd);
}


/* Print the average of the signal stored into "data"
   the "size" specifies the array size, "mess" is the prefix
   of the message to be printed on screen */
void average2d(_VALTYPE *data, int size, char *mess) {
  register int ix, iy;
  register double sum;
  sum = 0;
  for (ix = 0; ix < size; ix++) {
    for (iy = 0; iy < size; iy++) {
      sum += data[_XY(ix,iy,size)];
    }
  }
  printf("%s %e\n",mess,sum/(double)(size*size));
}

/* Initialize the initial 2D array with a tensorial
   product of sine function multiplied by a decreasing 
   exp(-x) coefficient in both directions x and y */
void init_func_modes(_VALTYPE *data, int size) {
  double  dpi, coeff, dx, val, maxval, inv_maxval;
  int     modex, ix, iy;
  double *tmpx, *tmpy;
  /* Two times PI */
  dpi   = 8*atan(1.);
  /* Excited mode */
  modex = MODEX;
  /* Mesh size */
  dx    = dpi/size;
  coeff = dx*modex;
  tmpx = (double *) malloc(size*sizeof(double));
  tmpy = (double *) malloc(size*sizeof(double));
  /* Initialize random generator */
  srand48(4);
  /* Fill "data" array with a signal + noise */
  /* The signal is a sine times an exponentional decrease */
  maxval = 0;
  for (iy = 0; iy < size; iy++) {
    tmpy[iy] = exp(COEFFEXP*dx*abs(iy-size/2));
    tmpy[iy] *= sin(coeff*iy);
  }
  for (ix = 0; ix < size; ix++) {
    tmpx[ix] = exp(COEFFEXP*dx*abs(ix-size/2));
    tmpx[ix] *= sin(coeff*ix);
  }
  for (iy = 0; iy < size; iy++) {
    for (ix = 0; ix < size; ix++) {
      //      val =  tmpx[ix]*tmpy[iy] + .002*drand48();
      val =  .01+tmpx[ix]*tmpy[iy];
      data[_XY(ix,iy,size)] = val;
      val = fabs(val);
      maxval = (val > maxval?val:maxval);
    }
  }
  inv_maxval = 1./maxval;
  for (iy = 0; iy < size; iy++) {
    for (ix = 0; ix < size; ix++) {
      data[_XY(ix,iy,size)] = data[_XY(ix,iy,size)] * inv_maxval;
    }
  }
}

/* Initialize the initial 2D array */
void init_func_stair(_VALTYPE *data, int size) {
  double  val, maxval, inv_maxval, dx, dy, dpi, tmpx;
  int     ix, iy, cy, cx, midsize = size/2;
  dpi   = 8*atan(1.);
  maxval = 0;
  dx = 2*dpi/size;
  dy = 2*dpi/size;
  printf("size %d ========\n",size);
  for (iy = 0; iy < size; iy++) {
    cy = iy - midsize;
    for (ix = 0; ix < size; ix++) {
      cx = ix - midsize;
      tmpx = sqrt((cx*dx)*(cx*dx)+(cy*dy)*(cy*dy));
      if (tmpx < MODEX) val = 1.;
      else val = 0;
      data[_XY(ix,iy,size)] = val; 
      val = fabs(val);
      maxval = (val > maxval?val:maxval);
    }
  }
  inv_maxval = 1./maxval;
  for (ix = 0; ix < size; ix++) {
    for (iy = 0; iy < size; iy++) {
      data[_XY(ix,iy,size)] = data[_XY(ix,iy,size)] * inv_maxval;
    }
  }
}

/* Initialize the initial 2D array */
void init_func_waterdrop(_VALTYPE *data, int size) {
  double  val, maxval, inv_maxval, tmpx, dx, dy;
  int     ix, iy, cx, cy, midsize = size/2;
  double  dpi;
  /* Two times PI */
  dpi   = 8*atan(1.);
  maxval = 0;
  dx = 2*dpi/size;
  dy = 2*dpi/size;
  for (iy = 0; iy < size; iy++) {
    cy = iy - midsize;
    for (ix = 0; ix < size; ix++) {
      cx = ix - midsize;
      tmpx = sqrt((cx*dx)*(cx*dx)+(cy*dy)*(cy*dy));
      val = cos(tmpx*MODEX)*exp(-tmpx*tmpx);
      data[_XY(ix,iy,size)] = val; 
      val = fabs(val);
      maxval = (val > maxval?val:maxval);
    }
  }
  inv_maxval = 1./maxval;
  for (ix = 0; ix < size; ix++) {
    for (iy = 0; iy < size; iy++) {
      data[_XY(ix,iy,size)] = data[_XY(ix,iy,size)] * inv_maxval;
    }
  }
}

//guint g_swap_hash(gconstpointer v);

/* Program that experiment a  wavelet scheme
   on a 2D data set */
int main(int argc, char **argv) {
  _VALTYPE *f_old, *f_wav;
  FTYPE    *s_wav;
  char     *c_grid;
  int      Nmax = TSIZE;
  double   normcoef = 1.;//pow(2,-1/Lp);
  double   thresholdMax = .003;
  double   dx;
  int      shift = 0;
  int      input;
  long long nbkb;
  //  char    message[50];

//  guint inp, out;
//  inp = 0x11UL;
//  out = g_swap_hash(&inp);
//  printf ("swap %x %x\n",inp,out);
//  printf ("sizeof guint %d\n",(int)sizeof(guint));
  f_old  = (_VALTYPE*)malloc(Nmax*Nmax*sizeof(_VALTYPE));
  c_grid = (char  *)malloc(Nmax*Nmax*sizeof(char));
  f_wav  = (_VALTYPE*)malloc(Nmax*Nmax*sizeof(_VALTYPE));
  dx    = 8*atan(1)/Nmax;
  if (argc > 1) input = atoi(argv[1]);
  else          input = 0;
  printf("Nmax = %d input = %d\n",Nmax,input);
  switch (input) {
  case 1:
    /* Initialize initial array with a staircase signal */
    GETTIME(init_func_stair(f_old,Nmax),"init_func:");
    break;
  case 2:
    /* Initialize initial array with a staircase signal */
    GETTIME(init_func_waterdrop(f_old,Nmax),"init_func:");
    break;
  default: 
    /* Initialize initial array with a modified sine signal */
    GETTIME(init_func_modes(f_old,Nmax),"init_func:");
  }
  /* print input signal into a file 'signal' */
  WAVDEBUG(print2d(f_old,Nmax,dx,"input.dat"));
  average2d(f_old,Nmax,"average of input:");

//  /*==========  UNITARY TEST0  ==========*/
//  thresholdMax = .05;
//  wav_init(15);
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Wavelet inverse transform */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,
//			   0.,c_grid,&shift),"threshold:");
//  WAVDEBUG(print2d(f_wav,Nmax,dx,"wav.dat"));
//  GETTIME(wav_itransform(f_wav,Nmax),"itransf:");
//  average2d(f_wav,Nmax,"average of output:");
//  WAVDEBUG(print2d(f_wav,Nmax,dx,"output.dat"));
//  exit(0);
//  /*=====================================*/

//  /*==========  UNITARY TEST1  ==========*/
//  wav_init(2);
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Thresholding of details */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,
//			   0.,c_grid,&shift),"threshold:");
//  /* Wavelet iverse transform */
//  WAVDEBUG(print2d(f_wav,Nmax,dx,"wav.dat"));
//  printf("average_deb of output: %e\n",f_wav[0]);
//  GETTIME(wav_itransform(f_wav,Nmax),"itransf:");
//  average2d(f_wav,Nmax,"average of input:");
//  WAVDEBUG(print2d(f_wav,Nmax,dx,"output.dat"));
//  exit(0);
//  /*=====================================*/

//  /*==========  UNITARY TEST2  ==========*/
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Thresholding of details */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,NOTAVALUE),"threshold:");
//  /* Copy signal from f_wav to s_wav */
//  GETTIME(wav_copyfunc_as(f_wav,s_s_wav,Nmax),"fcopy:");
//  exit(0);
//  /*=====================================*/

//  /*==========  UNITARY TEST3  ==========*/
//  wav_init(1);
//  /* Copy signal from f_old to f_wav */
//  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
//  average2d(f_wav,Nmax,"average of input:");
//  /* Wavelet forward transform */
//  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
//  /* Thresholding of details */
//  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,
//			   NOTAVALUE,c_grid),"threshold:");
//  /* Copy signal from f_wav to s_wav */
//  WAVDEBUG(print2d(f_wav,Nmax,dx,"wav.dat"));
//  GETTIME(wav_copyfunc_as(f_wav,s_wav,Nmax),"adapt_fcopy:");
//  WAVDEBUG(sparse_print2d(s_wav,Nmax,dx,"swav1.dat"));
//
//  GETTIME(wav_adapt_itransform(s_wav,Nmax),"adapt_itransf:");
//  WAVDEBUG(sparse_print2d(s_wav,Nmax,dx,"sparse.dat"));
//  GETTIME(wav_copyfunc_aa_select(f_wav,f_old,Nmax,c_grid,0.),
//	  "dense_fcopy:");
//  GETTIME(wav_itransform(f_old,Nmax),"itransf:");
//  average2d(f_old,Nmax,"average of output:");
//  WAVDEBUG(print2d(f_old,Nmax,dx,"output.dat"));
//  exit(0);
//  /*=====================================*/

  /*==========  UNITARY TEST4  ==========*/
  wav_init(2);
  /* Copy signal from f_old to f_wav */
  GETTIME(wav_copyfunc_aa(f_old,f_wav,Nmax),"fcopy_A:");
  average2d(f_wav,Nmax,"average of input:");
  /* Wavelet forward transform */
  GETTIME(wav_ftransform(f_wav,Nmax),"ftransf:");
  /* Thresholding of details */
  GETTIME(wav_thresholding(f_wav,Nmax,normcoef,thresholdMax,
			   NOTAVALUE,c_grid,&shift),"threshold:");
  /* Copy signal from f_wav to s_wav */
  get_mem_usage(&(vvms[0]), &(vrss[0]));
  _INITF (&s_wav,shift,Nmax);
  GETTIME(wav_copyfunc_as(f_wav,s_wav,Nmax),"adapt_fcopy:");
  wav_getkb(s_wav,Nmax);
  get_mem_usage(&(vvms[4]), &(vrss[4]));
  nbkb = vrss[4]-vrss[0];
  printf("Sparse Data size rss in KB: %ld\n",nbkb);
  printf("Sparse Data size rss in MB: %.2f\n",(1.*nbkb)/(1<<10));
  WAVDEBUG(sparse_print2d(s_wav,Nmax,dx,"sinit.dat"));

  wav_flushcache(16);
  GETTIME(wav_adapt_itransform(s_wav,Nmax),"adapt_itransf:");
  WAVDEBUG(sparse_print2d(s_wav,Nmax,dx,"sparse.dat"));
  GETTIME(wav_copyfunc_aa_select(f_wav,f_old,Nmax,c_grid,0.),
	  "dense_fcopy:");
  GETTIME(wav_itransform(f_old,Nmax),"itransf:");
  WAVDEBUG(print2d(f_old,Nmax,dx,"output.dat"));
  /*=====================================*/
  
  free(f_old);
  free(f_wav);
  free(c_grid);
  _DESTROY(&s_wav);
  return 0;
}
